import torch
import transformers
from transformers import BertTokenizer, BertForSequenceClassification
from torch.nn.parallel import DistributedDataParallel as DDP
import torch.distributed as dist
import time
from datetime import timedelta
import os
transformers.utils.logging.set_verbosity_error()

BATCH_SIZE = 2048
TRAINING_STEPS = 100

def init_process(backend='gloo'): # change to 'nccl' to use NCCL backend
    torch.cuda.set_device(int(os.environ['LOCAL_RANK']))
    dist.init_process_group(backend=backend, timeout=timedelta(seconds=60*60)) 

    rank = dist.get_rank()
    local_rank = int(os.environ['LOCAL_RANK'])
    world_size = dist.get_world_size()
    
    if rank == 0:
        num_cuda_devices = torch.cuda.device_count()
        print(f"Process {rank} initialized on local rank {local_rank} with backend {backend} and world_size {world_size}.")
        print(f"Number of locally available CUDA devices: {num_cuda_devices}")

def benchmark():
    model_name = "bert-base-uncased"
    tokenizer = BertTokenizer.from_pretrained(model_name)
    local_rank = int(os.environ['LOCAL_RANK'])
    world_size = dist.get_world_size()
    
    # Calculate per-GPU batch size
    per_gpu_batch_size = BATCH_SIZE // world_size
    
    model = BertForSequenceClassification.from_pretrained(model_name).cuda(local_rank)
    model = DDP(model, device_ids=[local_rank])
    optimizer = torch.optim.Adam(model.parameters())
    
    # Create different data for each GPU
    start_idx = local_rank * per_gpu_batch_size
    end_idx = start_idx + per_gpu_batch_size
    texts = [f"This is sample sentence {i} for benchmarking BERT." for i in range(start_idx, end_idx)]
    inputs = tokenizer(texts, return_tensors="pt", padding=True, truncation=True).to(local_rank)
    labels = torch.ones(per_gpu_batch_size, dtype=torch.long).to(local_rank)
    
    dist.barrier()
    start_time = time.time()
    
    for _ in range(TRAINING_STEPS):
        optimizer.zero_grad()
        outputs = model(**inputs, labels=labels)
        loss = outputs.loss
        loss.backward()
        optimizer.step()
    
    dist.barrier()
    end_time = time.time()

    if dist.get_rank() == 0:
        print(f"Time taken for {TRAINING_STEPS} forward and backward pass(es) with BATCH_SIZE={BATCH_SIZE}: {end_time - start_time} seconds")
    dist.destroy_process_group()

if __name__ == "__main__":
    init_process()
    benchmark()